use std::{
    io::{stdin, stdout, Write},
    path::PathBuf,
};

use anyhow::{anyhow, bail, Result};
use clap::{Args, Parser, Subcommand};
use id3::{Tag, TagLike};
use musixmatch_inofficial::{
    models::{SubtitleFormat, Track, TrackId, TranslationMap},
    Musixmatch,
};

#[derive(Parser)]
#[command(author, version, about, long_about = None)]
struct Cli {
    #[command(subcommand)]
    command: Commands,
}

#[derive(Subcommand)]
enum Commands {
    Get {
        #[command(subcommand)]
        command: GetCommands,
    },
    Mp3 {
        #[command(subcommand)]
        command: FileCommands,
    },
    #[group(required = true)]
    Search {
        /// Track name
        #[clap(short, long)]
        name: Option<String>,
        /// Artist
        #[clap(short, long)]
        artist: Option<String>,
        /// Lyrics
        #[clap(short, long)]
        lyrics: Option<String>,
        /// Search query
        query: Option<Vec<String>>,
    },
}

#[derive(Subcommand)]
enum GetCommands {
    Lyrics {
        #[clap(flatten)]
        ident: TrackIdentifiers,
        /// Language
        #[clap(long)]
        lang: Option<String>,
        /// Bilingual
        #[clap(long)]
        bi: bool,
    },
    Subtitles {
        #[clap(flatten)]
        ident: TrackIdentifiers,
        /// Track length
        #[clap(long, short)]
        length: Option<f32>,
        /// Maximum deviation from track length (Default: 1s)
        #[clap(long)]
        max_deviation: Option<f32>,
        /// Subtitle format
        #[clap(long, default_value = "lrc")]
        format: SubtitleFormatClap,
        /// Language
        #[clap(long)]
        lang: Option<String>,
    },
    Track {
        #[clap(flatten)]
        ident: TrackIdentifiers,
    },
}

#[derive(Args)]
struct TrackIdentifiers {
    /// Track name
    #[clap(short, long)]
    name: Option<String>,
    /// Artist
    #[clap(short, long)]
    artist: Option<String>,
    /// Musixmatch-ID
    #[clap(long)]
    mxm_id: Option<u64>,
    /// Musixmatch-Commontrack-ID
    #[clap(long)]
    mxm_ct: Option<u64>,
    /// Musixmatch-Vanity-ID
    #[clap(long)]
    mxm_vanity: Option<String>,
    /// Spotify-ID
    #[clap(long)]
    spotify: Option<String>,
    /// Musicbrainz-ID
    #[clap(long)]
    musicbrainz: Option<String>,
    /// ISRC
    #[clap(long)]
    isrc: Option<String>,
}

#[derive(Subcommand)]
enum FileCommands {
    Lyrics {
        /// Music file
        #[clap(value_parser)]
        file: PathBuf,
    },
    Subtitles {
        /// Music file
        #[clap(value_parser)]
        file: PathBuf,
    },
}

#[derive(clap::ValueEnum, Debug, Copy, Clone)]
pub enum SubtitleFormatClap {
    Lrc,
    Ttml,
    TtmlStructured,
    Json,
    EbuStl,
}

impl From<SubtitleFormatClap> for SubtitleFormat {
    fn from(format: SubtitleFormatClap) -> Self {
        match format {
            SubtitleFormatClap::Lrc => SubtitleFormat::Lrc,
            SubtitleFormatClap::Ttml => SubtitleFormat::Ttml,
            SubtitleFormatClap::TtmlStructured => SubtitleFormat::TtmlStructured,
            SubtitleFormatClap::Json => SubtitleFormat::Json,
            SubtitleFormatClap::EbuStl => SubtitleFormat::EbuStl,
        }
    }
}

const NA_STR: &str = "n/a";

#[tokio::main]
async fn main() {
    let cli = Cli::parse();

    run(cli).await.unwrap_or_else(|e| {
        println!("Error: {}", e);
        std::process::exit(1);
    });
}

async fn run(cli: Cli) -> Result<()> {
    let mut storage_file = dirs::data_dir().ok_or(anyhow!("no data dir"))?;
    storage_file.push("musixmatch-cli");
    std::fs::create_dir_all(&storage_file)?;
    storage_file.push("musixmatch_session.json");

    let mxm = Musixmatch::builder().storage_file(storage_file).build()?;

    match mxm.login().await {
        Ok(_) => {}
        Err(musixmatch_inofficial::Error::MissingCredentials) => {
            println!("Enter your Musixmatch credentials:");
            let email = input("E-Mail: ");
            let password = input_pwd("Password: ");

            mxm.set_credentials(&email, &password);
            mxm.login().await?;
        }
        Err(e) => bail!("login error: {}", e),
    };

    match cli.command {
        Commands::Get { command } => match command {
            GetCommands::Lyrics { ident, lang, bi } => {
                let track_id = get_track_id(ident, &mxm).await?;
                let lyrics = mxm.track_lyrics(track_id.clone()).await?;

                eprintln!("Lyrics ID: {}", lyrics.lyrics_id);
                eprintln!(
                    "Language: {}",
                    lyrics.lyrics_language.as_deref().unwrap_or(NA_STR)
                );
                eprintln!(
                    "Copyright: {}",
                    lyrics
                        .lyrics_copyright
                        .as_deref()
                        .map(|c| c.trim())
                        .unwrap_or(NA_STR)
                );

                let mut lyrics_body = lyrics.lyrics_body;

                if let Some(lang) = lang {
                    if Some(&lang) != lyrics.lyrics_language.as_ref() {
                        let tl = mxm.track_lyrics_translation(track_id, &lang).await?;
                        if tl.is_empty() {
                            eprintln!(
                                "Translation not found. Returning lyrics in original language."
                            );
                        } else {
                            eprintln!("Translated to: {}", tl.lang);
                            let tm = TranslationMap::from(tl);
                            let translated = tm.translate_lyrics(&lyrics_body);
                            lyrics_body = if bi {
                                lyrics_body
                                    .lines()
                                    .zip(translated.lines())
                                    .map(|(a, b)| {
                                        if a == b {
                                            a.to_string() + "\n"
                                        } else {
                                            format!("{a}\n> {b}\n")
                                        }
                                    })
                                    .collect()
                            } else {
                                translated
                            };
                        }
                    }
                }

                eprintln!();
                println!("{}", lyrics_body);
            }
            GetCommands::Subtitles {
                ident,
                length,
                max_deviation,
                format,
                lang,
            } => {
                let track_id = get_track_id(ident, &mxm).await?;
                let subtitles = mxm
                    .track_subtitle(
                        track_id.clone(),
                        if lang.is_some() {
                            SubtitleFormat::Json
                        } else {
                            format.into()
                        },
                        length,
                        max_deviation.or(Some(1.0)),
                    )
                    .await?;

                eprintln!("Subtitle ID: {}", subtitles.subtitle_id);
                eprintln!(
                    "Language: {}",
                    subtitles.subtitle_language.as_deref().unwrap_or(NA_STR)
                );
                eprintln!("Length: {}", subtitles.subtitle_length);
                eprintln!(
                    "Copyright: {}",
                    subtitles
                        .lyrics_copyright
                        .as_deref()
                        .map(|s| s.trim())
                        .unwrap_or(NA_STR)
                );

                if let Some(lang) = lang {
                    let mut lines = subtitles.to_lines()?;

                    if Some(&lang) != subtitles.subtitle_language.as_ref() {
                        let tl = mxm.track_lyrics_translation(track_id, &lang).await?;
                        if tl.is_empty() {
                            bail!("Translation not found")
                        } else {
                            eprintln!("Translated to: {}", tl.lang);
                            let tm = TranslationMap::from(tl);
                            lines = tm.translate_subtitles(&lines);
                        }
                    }

                    eprintln!();
                    let res = match format {
                        SubtitleFormatClap::Lrc => lines.to_lrc(),
                        SubtitleFormatClap::Ttml => lines.to_ttml(),
                        SubtitleFormatClap::Json => lines.to_json()?,
                        SubtitleFormatClap::TtmlStructured | SubtitleFormatClap::EbuStl => {
                            bail!("subtitle format {format:?} cant be translated")
                        }
                    };
                    println!("{}", res);
                } else {
                    eprintln!();
                    println!("{}", subtitles.subtitle_body);
                }
            }
            GetCommands::Track { ident } => {
                let track = get_track(ident, &mxm).await?;
                println!("{}", serde_json::to_string_pretty(&track)?)
            }
        },
        Commands::Mp3 { command } => match command {
            FileCommands::Lyrics { file } => {
                let tag = Tag::read_from_path(&file)?;

                let title = tag.title().ok_or(anyhow!("no title"))?;
                let artist = tag.artist().ok_or(anyhow!("no artist"))?;

                let lyrics = mxm.matcher_lyrics(title, artist).await?;

                println!(
                    "Lyrics for {} by {}:\n\n{}",
                    title, artist, lyrics.lyrics_body
                );
            }
            FileCommands::Subtitles { file } => {
                let tag = Tag::read_from_path(&file)?;
                let duration = mp3_duration::from_path(&file)?;

                let title = tag.title().ok_or(anyhow!("no title"))?;
                let artist = tag.artist().ok_or(anyhow!("no artist"))?;

                let subtitles = mxm
                    .matcher_subtitle(
                        title,
                        artist,
                        SubtitleFormat::Lrc,
                        Some(duration.as_secs_f32()),
                        Some(1.0),
                    )
                    .await?;

                println!("{}", subtitles.subtitle_body);
            }
        },
        Commands::Search {
            query,
            name,
            artist,
            lyrics,
        } => {
            let mut sb = mxm
                .track_search()
                .s_track_rating(musixmatch_inofficial::models::SortOrder::Desc);
            let querystr;
            if let Some(q) = &query {
                querystr = q.join(" ");
                sb = sb.q(&querystr);
            }
            if let Some(n) = &name {
                sb = sb.q_track(n);
            }
            if let Some(a) = &artist {
                sb = sb.q_artist(a);
            }
            if let Some(l) = &lyrics {
                sb = sb.q_lyrics(l);
            }

            let tracks = sb.send(20, 0).await?;
            for t in tracks {
                println!(
                    "{} - {} ({})  ISRC'{}' <https://musixmatch.com/lyrics/{}>",
                    t.track_name,
                    t.artist_name,
                    t.first_release_date.map(|d| d.year()).unwrap_or_default(),
                    t.track_isrc.unwrap_or_default(),
                    t.commontrack_vanity_id
                );
            }
        }
    };
    Ok(())
}

enum TrackOrId<'a> {
    Track(Box<Track>),
    TrackId(TrackId<'a>),
}

async fn get_track_or_id(
    ident: TrackIdentifiers,
    mxm: &Musixmatch,
    translation_status: bool,
) -> Result<TrackOrId<'static>> {
    Ok(
        match (
            ident.name,
            ident.artist,
            ident.mxm_id,
            ident.mxm_ct,
            ident.mxm_vanity,
            ident.spotify,
            ident.musicbrainz,
            ident.isrc,
        ) {
            (_, _, Some(mxm_id), _, _, _, _, _) => TrackOrId::TrackId(TrackId::TrackId(mxm_id)),
            (_, _, _, Some(mxm_ct), _, _, _, _) => TrackOrId::TrackId(TrackId::Commontrack(mxm_ct)),
            (_, _, _, _, Some(mxm_vanity), _, _, _) => {
                TrackOrId::TrackId(TrackId::CommontrackVanity(mxm_vanity.into()))
            }
            (_, _, _, _, _, Some(spotify), _, _) => {
                TrackOrId::TrackId(TrackId::Spotify(spotify.into()))
            }
            (_, _, _, _, _, _, Some(musicbrainz), _) => {
                TrackOrId::TrackId(TrackId::Musicbrainz(musicbrainz.into()))
            }
            (_, _, _, _, _, _, _, Some(isrc)) => TrackOrId::TrackId(TrackId::Isrc(isrc.into())),
            (Some(name), Some(artist), _, _, _, _, _, _) => TrackOrId::Track(Box::new(
                mxm.matcher_track(&name, &artist, "", translation_status, true)
                    .await?,
            )),
            _ => bail!("no track identifier given"),
        },
    )
}

async fn get_track_id(ident: TrackIdentifiers, mxm: &Musixmatch) -> Result<TrackId<'static>> {
    Ok(match get_track_or_id(ident, mxm, false).await? {
        TrackOrId::Track(track) => TrackId::TrackId(track.track_id),
        TrackOrId::TrackId(id) => id,
    })
}

async fn get_track(ident: TrackIdentifiers, mxm: &Musixmatch) -> Result<Track> {
    Ok(match get_track_or_id(ident, mxm, true).await? {
        TrackOrId::Track(track) => *track,
        TrackOrId::TrackId(id) => mxm.track(id, true, true).await?,
    })
}

fn input(prompt: &str) -> String {
    print!("{}", prompt);

    stdout().flush().expect("Failed to flush stdout!");

    let mut input = String::new();
    stdin().read_line(&mut input).expect("Failed to read line");
    // Remove trailing newline
    input.pop();
    input
}

fn input_pwd(prompt: &str) -> String {
    print!("{}", prompt);
    stdout().flush().expect("Failed to flush stdout!");

    rpassword::read_password().expect("Failed to read password")
}
